import magicdb
from magicdb.database import db
from magicdb.utils.Serverless.span import safe_span
import threading
from pydantic import BaseModel
from enum import Enum
from typing import List, Dict, Tuple, Any

MAX_BATCH_SIZE = 400


class OperationType(str, Enum):
    SET = "set"
    UPDATE = "update"
    DELETE = "delete"


class BatchCommand(BaseModel):
    operation_type: OperationType
    args: tuple = ()
    kwargs: dict = {}


class BatchWrapper:
    def __init__(self, *args, max_batch_size=MAX_BATCH_SIZE, **kwargs):
        self.max_batch_size = max_batch_size
        if self.max_batch_size != MAX_BATCH_SIZE:
            print(f"{self.max_batch_size=}")
        self._init_args: tuple = args
        self._init_kwargs: dict = kwargs
        self._batch_commands: List[BatchCommand] = []

    def set(self, *args, **kwargs):
        self._batch_commands.append(
            BatchCommand(operation_type=OperationType.SET, args=args, kwargs=kwargs)
        )

    def update(self, *args, **kwargs):
        self._batch_commands.append(
            BatchCommand(operation_type=OperationType.UPDATE, args=args, kwargs=kwargs)
        )

    def delete(self, *args, **kwargs):
        self._batch_commands.append(
            BatchCommand(operation_type=OperationType.DELETE, args=args, kwargs=kwargs)
        )

    @staticmethod
    def commit_batch(batch, i=None):
        with safe_span("batch_commit"):
            batch.commit()
            print(f"Batch number {'?' if i is None else i} committed!")

    def commit_batch_async(self, batch, i=None):
        t = threading.Thread(target=self.commit_batch, args=(batch, i))
        t.start()
        return t

    def commit_batches(self, batches: List[magicdb.WriteBatch], sync=False):
        if not sync:
            ts = [self.commit_batch_async(batch, i) for i, batch in enumerate(batches)]
            [t.join() for t in ts]
        else:
            [self.commit_batch(batch, i) for i, batch in enumerate(batches)]

    def commit(self, sync=False):
        batches: List[magicdb.WriteBatch] = self.make_batches()
        self.commit_batches(batches=batches, sync=sync)

    def commit_async(self, *args, **kwargs):
        t = threading.Thread(target=self.commit, args=args, kwargs=kwargs)
        t.start()
        return t

    def make_batches(self) -> List[magicdb.WriteBatch]:
        chunks = self.chunk_array(
            arr=self._batch_commands, chunk_size=self.max_batch_size
        )
        batches: List[magicdb.WriteBatch] = []
        for i, chunk in enumerate(chunks):
            batch: magicdb.WriteBatch = db.conn.batch(
                *self._init_args, **self._init_kwargs
            )
            for batch_command in chunk:
                batch_command: BatchCommand
                if batch_command.operation_type == OperationType.SET:
                    batch.set(*batch_command.args, **batch_command.kwargs)
                elif batch_command.operation_type == OperationType.UPDATE:
                    batch.update(*batch_command.args, **batch_command.kwargs)
                elif batch_command.operation_type == OperationType.DELETE:
                    batch.delete(*batch_command.args, **batch_command.kwargs)

            batches.append(batch)
        return batches

    @staticmethod
    def chunk_array(arr, chunk_size):
        return [arr[i : i + chunk_size] for i in range(0, len(arr), chunk_size)]



from magicdb import db

"""MAGIC MODEL MAGIC FIELDS HELPERS"""


def make_magic_fields_from_kwargs(kwargs: dict, collection_name: str) -> None:
    """Will make magic_fields either from key, ref, or id (included no id given or parent given).
    Whichever one exists, in that order."""
    if kwargs.get("key", None):
        make_magic_fields_from_key(kwargs)
    elif kwargs.get("ref", None):
        make_magic_fields_from_ref(kwargs)
    else:
        make_magic_fields_from_id_parent_or_nothing(kwargs, collection_name)


def make_magic_fields_from_key(kwargs):
    kwargs["ref"] = db.conn.document(kwargs["key"])
    kwargs["id"] = id_from_ref(kwargs["ref"])


def make_magic_fields_from_ref(kwargs):
    kwargs["key"] = key_from_ref(kwargs["ref"])
    kwargs["id"] = id_from_ref(kwargs["ref"])


def make_magic_fields_from_id_parent_or_nothing(kwargs, collection_name):
    """Assigns db_fields fields using an id if given. Otherwise, makes an id... Also uses parent if necessary."""
    collection_ref = (
        db.conn.collection(collection_name)
        if not kwargs.get("parent")
        else kwargs.get("parent").ref.collection(collection_name)
    )
    kwargs["ref"] = (
        collection_ref.document()
        if not kwargs.get("id")
        else collection_ref.document(kwargs.get("id"))
    )
    make_magic_fields_from_ref(kwargs)


def id_from_ref(ref):
    return ref.path.split("/")[-1]


def key_from_ref(ref):
    return ref.path



from __future__ import annotations
from typing import List, Dict, Any, ClassVar, Set
import typing as T
from datetime import datetime
import time
from copy import deepcopy
import json
import os
import pprint
import uuid
from glom import glom

from magicdb.Timing.decorator import timing


from pydantic import BaseModel, Field, PrivateAttr
from pydantic.main import ModelMetaclass

import magicdb
from magicdb.Queries import QuerySpeed
from magicdb.utils.Serverless.span import safe_span

from magicdb.utils.updating_objects import make_update_obj
from magicdb.utils.async_helpers import magic_async

from magicdb.Models.magic_fields_maker import FieldMaker

from magicdb.Models.MagicCache import MagicCache

import redis


class DatabaseError(Exception):
    def __init__(self, key):
        self.message = (
            f"There is no document with key {key} to update. Add update.(create=True) to save the document"
            f" if it does not exist. Otherwise, you can save the document: save()."
        )


class QueryMeta(type):
    """https://stackoverflow.com/questions/128573/using-property-on-classmethods"""

    @property
    def collection(cls) -> QuerySpeed:
        return QuerySpeed(cls)

    @property
    def collection_group(cls) -> QuerySpeed:
        return QuerySpeed(cls).collection_group()


class QueryAndBaseMetaClasses(ModelMetaclass, QueryMeta):
    pass


class MagicModelSpeed(BaseModel, metaclass=QueryAndBaseMetaClasses):
    """
    When this gets inited, if given an id or key, assign based on that.
    Otherwise, assign them based on what Firestore gives it
    """

    id: str = Field(default_factory=uuid.uuid4)
    _key: str = PrivateAttr(None)
    _ref: magicdb.DocumentReference = PrivateAttr(None)
    _doc: magicdb.DocumentSnapshot = PrivateAttr(None)
    _parent: MagicModelSpeed = PrivateAttr(None)
    _data_from_db: dict = PrivateAttr(None)

    _magic_fields_setters: Set[str] = PrivateAttr(default_factory=set)
    _magic_fields: Set[str] = PrivateAttr(default_factory=set)

    __call__ = ...  # to satisfy Query python linter

    @timing
    def __init__(self, from_db: bool = False, **data):
        data_from_db = deepcopy(data)
        super().__init__(**data)

        FieldMaker(
            **data, _collection_name=self.get_collection_name()
        ).set_model_fields(self)

        self._data_from_db = {} if not from_db else data_from_db

        self._magic_fields: Set[str] = MagicModelSpeed.get_magic_fields()
        self._magic_fields_setters: Set[str] = MagicModelSpeed.get_magic_set_fields()

    @classmethod
    @timing
    def construct(cls, from_db: bool = False, **data):
        data_from_db = deepcopy(data)
        new_obj = super().construct(**data)

        FieldMaker(
            **data, _collection_name=new_obj.get_collection_name()
        ).set_model_fields(new_obj)

        new_obj._data_from_db = {} if not from_db else data_from_db
        return new_obj

    """GETTING AND SETTING FIELDS"""

    def set_id(self, id: str, fields_to_exclude: Set[str] = None):
        FieldMaker(
            id=id, _parent=self._parent, _collection_name=self.get_collection_name()
        ).set_model_fields(self, fields_to_exclude=fields_to_exclude)

    """Must make this because id is an actual field."""

    def set_key(self, key: str, fields_to_exclude: Set[str] = None):
        FieldMaker(
            _key=key, _collection_name=self.get_collection_name()
        ).set_model_fields(self, fields_to_exclude=fields_to_exclude)

    @property
    def key(self):
        return self._key

    @key.setter
    def key(self, key: str):
        self.set_key(key=key)

    def set_ref(
        self, ref: magicdb.DocumentReference, fields_to_exclude: Set[str] = None
    ):
        FieldMaker(
            _ref=ref, _collection_name=self.get_collection_name()
        ).set_model_fields(self, fields_to_exclude=fields_to_exclude)

    @property
    def ref(self):
        return self._ref

    @ref.setter
    def ref(self, ref: magicdb.DocumentReference):
        self.set_ref(ref=ref)

    def set_parent(self, parent: MagicModelSpeed, fields_to_exclude: Set[str] = None):
        print("SETTING PARENT", parent, "fields to exclude", fields_to_exclude)
        FieldMaker(
            _parent=parent, id=self.id, _collection_name=self.get_collection_name()
        ).set_model_fields(self, fields_to_exclude=fields_to_exclude)

    @property
    def parent(self):
        return self._parent

    @parent.setter
    def parent(self, parent: MagicModelSpeed):
        self.set_parent(parent=parent)

    def __setattr__(self, name, value):
        if name in self._magic_fields_setters:
            if func := getattr(self, f"set_{name}"):
                func(value, fields_to_exclude={name})
            if f"_{name}" in self._magic_fields:  # for id
                name = f"_{name}"
        """ This is for when you only want to edit the id with nothing happening."""
        if name == "_id":
            name = "id"
        super().__setattr__(name, value)

    """PRINTING AND RETURNING"""

    def __repr__(self, *args, **kwargs):
        return f"{self.__class__.__name__}({self.__repr_str__(', ')})"

    def __str__(self, *args, **kwargs):
        return f"{self.__repr_str__(' ')}"

    def __repr_str__(self, join_str=", "):
        key_values: List[str] = []
        for field, val in self.__dict__.items():
            key_values.append(f"{repr(field)}={repr(val)}")
        return join_str.join(key_values)

    def print_magic_fields(self, join_str=" "):
        key_values = []
        for field in MagicModelSpeed.get_magic_fields():
            key_values.append(f"{repr(field)}={repr(getattr(self, field, None))}")
        print(join_str.join(key_values))

    """META"""

    class Meta:
        """Init the meta class so you can use it and know it is there"""

        """Looks like having Meta in the TestModel actually overwrites this one, not inherits it"""
        ...

    @property
    def collection_name(self) -> str:
        return self.get_collection_name()

    @classmethod
    def make_default_collection_name(cls) -> str:
        return cls.__name__.lower()

    @classmethod
    def get_collection_name(cls) -> str:
        return getattr(cls.Meta, "collection_name", cls.make_default_collection_name())

    @classmethod
    def get_redis(cls) -> T.Optional[redis.Redis]:
        return getattr(cls.Meta, "redis", None)

    @classmethod
    def get_redis_ttl_secs(cls) -> T.Optional[int]:
        return getattr(cls.Meta, "redis_ttl_secs", None)

    """COMPARISONS"""

    def equal(self, model: BaseModel, path_to_compare=None):
        """The difference between objects will be firestore adds nanoseconds to the object. So parse those out"""
        replace_text = "+00:00"
        self_json = self.json().replace(replace_text, "")
        other_json = model.json().replace(replace_text, "")
        if path_to_compare:
            self_json = json.dumps(
                glom(json.loads(self_json), ".".join(path_to_compare))
            )
            other_json = json.dumps(
                glom(json.loads(other_json), ".".join(path_to_compare))
            )
        return self_json == other_json

    def been_updated(self, path_to_compare=None):
        from_db = self.__class__.get_with_error(self.id)
        return not self.equal(from_db, path_to_compare=path_to_compare)

    class Serverless:
        """Fills this with the latest context if it exists"""

        context = None

    class Config:
        anystr_strip_whitespace: bool = True
        arbitrary_types_allowed: bool = True
        json_encoders = {magicdb.DocumentReference: lambda doc_ref: doc_ref.path}

    """ ADDING TO FIRESTORE """

    @staticmethod
    def get_magic_fields() -> Set[str]:
        return {"id", *MagicModelSpeed.__private_attributes__.keys()}

    @staticmethod
    def get_magic_set_fields() -> Set[str]:
        s = set()
        for f in MagicModelSpeed.get_magic_fields():
            if f[0] == "_":
                s.add(f[1:])
            else:
                s.add(f)
        return s

    @staticmethod
    def remove_magic_fields(og_d) -> dict:
        new_d = deepcopy(og_d)
        for field in MagicModelSpeed.get_magic_fields():
            if field in new_d:
                del new_d[field]
        return new_d

    @timing
    def validate_all_fields(self):
        valid_obj_no_magic = self.__class__(**self.dict())
        self.__dict__.update(valid_obj_no_magic.__dict__)

    @magic_async
    @timing
    def save(self, batch=None, transaction=None, merge=False, ignore_fields=False):
        """Will create a new obj_to_save and save it so that all of the validation happens properly on a new obj."""
        batch_or_transaction = batch or transaction
        # make sure all the fields are correct by validating again
        if not ignore_fields:
            self.validate_all_fields()

        new_d = self.dict()
        # just add exclude = magic fields... also why make a deep copy? And you can do that anyway
        new_d = self.remove_magic_fields(new_d)  # must remove id

        with safe_span(f"save-{self._key}", use=(batch_or_transaction is None)):
            if batch_or_transaction:
                res = batch_or_transaction.set(self._ref, new_d, merge=merge)
            else:
                res = self._ref.set(new_d, merge=merge)

        if not merge:
            self._data_from_db = deepcopy(new_d)

        # update self just in case
        return self

    @magic_async
    @timing
    def update(
        self,
        batch=None,
        transaction=None,
        create=False,
        ignore_fields=False,
        print_update_d=False,
        only_print_update_d=False,
    ):
        batch_or_transaction = batch or transaction
        if not ignore_fields:
            self.validate_all_fields()

        new_d = self.dict()
        new_d = self.remove_magic_fields(new_d)  # must remove id
        data_from_db = self.remove_magic_fields(self._data_from_db)

        if not data_from_db:
            update_d = new_d
        else:
            update_d = make_update_obj(original=data_from_db, new=new_d)

        if print_update_d or only_print_update_d:
            pprint.pprint({f"update_d_for_{self.id}": update_d})
        if only_print_update_d:
            return
        try:
            if update_d != {}:
                with safe_span(
                    f"update-{self._key}", use=(batch_or_transaction is None)
                ):
                    self._ref.update(
                        update_d
                    ) if not batch_or_transaction else batch_or_transaction.update(
                        self._ref, update_d
                    )
            else:
                print(f"update_d was empty so did not update for {self.id}.")
            return self
        except Exception as e:
            if hasattr(e, "message") and "no document to update" in e.message.lower():
                if create:
                    return self.save(batch=batch, transaction=transaction)
                else:
                    db_error = DatabaseError(self._key)
                    raise DatabaseError(db_error.message)
            raise e

    @magic_async
    @timing
    def delete(
        self,
        batch=None,
        transaction=None,
        print_update_d=False,
        only_print_update_d=False,
    ):
        if print_update_d or only_print_update_d:
            print(f"About to delete {self.id}.")
        if only_print_update_d:
            return

        batch_or_transaction = batch or transaction
        with safe_span(f"delete-{self._key}", use=(batch_or_transaction is None)):
            return (
                self._ref.delete()
                if not batch_or_transaction
                else batch_or_transaction.delete(self._ref)
            )

    """QUERYING AND COLLECTIONS"""

    def exists(self):
        return self.__class__.collection.get(self.id) is not None

    def get_subcollections(self):
        return list(self.__class__.collection.document(self.id).collections())

    @classmethod
    def get_ref(cls, id) -> magicdb.DocumentReference:
        temp_cls = cls.construct(id=id)
        return temp_cls._ref

    @classmethod
    def get_default_exception(cls):
        return getattr(cls.Meta, "default_exception", Exception)

    @classmethod
    @magic_async
    def get_with_error(
        cls,
        id,
        error=None,
        cache: MagicCache = None,
        **kwargs,
    ):
        if cache is not None:
            if val := cache.get(cls=cls, id=id):
                print("cache hit!")
                return val
        if not error:
            error = cls.get_default_exception()
        val = cls.collection.get(id, **kwargs)
        if not val:
            raise error(f"{cls.__name__} with id {id} does not exist.")
        if cache is not None:
            cache.add(cls=cls, id=id, val=val)
        return val

    """GETTING SUBCLASSES"""

    @classmethod
    def get_subclasses(cls):
        all_subs = []
        for sub in cls.__subclasses__():
            all_subs.append(sub)
            all_subs += sub.get_subclasses()
        return list(set(all_subs))

    @staticmethod
    def get_all_subclasses_of_model():
        all_subs = []
        for sub in list(MagicModelSpeed.__subclasses__()):
            all_subs.append(sub)
            all_subs += sub.get_subclasses()
        return list(set(all_subs))

    @classmethod
    @magic_async
    def get_by_field(cls, field_name: str, value):
        models = cls.collection.where(
            magicdb.db.conn.field_path(field_name), "==", value
        ).stream()
        return None if not models else models[0]

    @classmethod
    @magic_async
    def get_by_field_with_error(cls, field_name: str, value, error=None):
        if not error:
            error = cls.get_default_exception()
        model = cls.get_by_field(field_name=field_name, value=value)
        if not model:
            raise error(f"{cls.__name__} with {field_name} {value} does not exist.")
        return model

    """Save history to DB"""

    @magic_async
    def save_history(self, user_id: str = None, **kwargs):
        """Saves current object to a changes sub collection"""
        history_key = f"{self._key}/history/{str(int(time.time()))}"

        class History(self.__class__):
            changed_at: datetime = Field(default_factory=datetime.utcnow)
            changed_by: str = user_id

        history_obj = History(**self.dict(), _key=history_key)
        return history_obj.save(**kwargs)

    @magic_async
    def save_deleted(self, user_id: str = None, **kwargs):
        deleted_key = f"deleted_{self.collection_name}/{self.id}"

        class Deleted(self.__class__):
            deleted_at: datetime = Field(default_factory=datetime.utcnow)
            deleted_by: str = user_id

        deleted_obj = Deleted(**self.dict(), _key=deleted_key)
        return deleted_obj.save(**kwargs)



from typing import ClassVar
from pydantic import BaseModel, validator
from slugify import slugify as base_slugify


class SlugModelSpeed(BaseModel):
    """ Adds slug and helper slug methods to MagicModel. Must use this with MagicModel """

    _slugify_kwargs: ClassVar[dict] = {"separator": "", "replacements": [["&", "and"]]}
    slug: str

    @classmethod
    def slugify(cls, slug):
        return base_slugify(slug, **cls._slugify_kwargs)

    @classmethod
    def create_slug(cls, value_to_slugify: str):
        slug_base = cls.slugify(value_to_slugify)
        slug = slug_base
        number = 1
        while cls.get_by_slug(slug):
            slug = cls.slugify(f"{slug_base}{number}")
            number += 1
        return slug

    @classmethod
    def get_by_slug(cls, slug):
        models = cls.collection.where("slug", "==", slug).stream()
        return None if not models else models[0]

    def unique_slug(self, slug):
        models = self.get_by_slug(slug)
        for model in models:
            if model.id != self.id:
                return False
        return True


class SlugModel(SlugModelSpeed):
    @validator("slug")
    def parse_slug(cls, v):
        return cls.slugify(v)



from typing import Dict, Any, Union


class MagicCache:
    def __init__(self):
        self._class_cache: [Dict[str, Dict[str, Any]]] = {}

    def get(self, cls: type, id: str) -> Union[Any, None]:
        if cls.__name__ not in self._class_cache:
            return None
        return self._class_cache[cls.__name__].get(id, None)

    def add(self, cls: type, id: str, val: Any) -> None:
        if cls.__name__ not in self._class_cache:
            self._class_cache[cls.__name__] = {}
        self._class_cache[cls.__name__][id] = val

    def remove(self, cls: type, id: str) -> bool:
        if cls.__name__ not in self._class_cache:
            return False
        if id not in self._class_cache[cls.__name__]:
            return False
        del self._class_cache[cls.__name__][id]
        return True

    def __str__(self):
        return str(self.__repr__())

    def __repr__(self):
        return self._class_cache



import typing as T
from typing import Union, List
from magicdb import db, ASCENDING, DESCENDING, DocumentSnapshot
from magicdb.Queries.Q import Q
from magicdb.utils.Serverless.span import safe_span
from magicdb.utils.async_helpers import magic_async

import redis as redis_lib
import time
import ujson


class QuerySpeed:
    def __init__(self, cls):
        from magicdb.Models.MagicModelSpeed import MagicModelSpeed

        self.cls: MagicModelSpeed = cls
        self.collection_path = self.cls.get_collection_name()
        self.firebase_query = self.create_query()
        self.query_inputs: List[str] = []

    @property
    def redis(self) -> T.Optional[redis_lib.Redis]:
        return self.cls.get_redis()

    @property
    def redis_ttl_secs(self) -> T.Optional[int]:
        return self.cls.get_redis_ttl_secs()

    def create_query(self):
        self.firebase_query = db.conn.collection(self.collection_path)
        return self.firebase_query

    def parent(self, parent_or_path):
        from magicdb.Models.MagicModelSpeed import MagicModelSpeed

        parent_path = (
            parent_or_path
            if not issubclass(type(parent_or_path), MagicModelSpeed)
            else parent_or_path._key
        )

        self.collection_path = parent_path + "/" + self.collection_path
        self.create_query()
        return self

    def document(self, document_path):
        self.firebase_query = self.firebase_query.document(document_path)
        return self

    def collection_group(self):
        self.firebase_query = db.conn.collection_group(self.cls.get_collection_name())
        return self

    def collection(self, collection_path):
        self.firebase_query = self.firebase_query.collection(collection_path)
        return self

    """Wrapping Firestore abilities."""

    def where(self, field, action, value):
        self.firebase_query = self.firebase_query.where(field, action, value)
        self.query_inputs.append(f"{field}, {action}, {value}")
        return self

    def order_by(self, field, direction=None, **kwargs):
        if direction:
            kwargs["direction"] = direction
        if kwargs.get("direction", None) == "asc":
            kwargs["direction"] = ASCENDING
        if kwargs.get("direction", None) == "desc":
            kwargs["direction"] = DESCENDING
        self.firebase_query = self.firebase_query.order_by(field, **kwargs)
        return self

    def start_at(self, fields):
        self.firebase_query = self.firebase_query.start_at(fields)
        return self

    def start_after_id(self, last_doc_id: str = None):
        if not last_doc_id:
            return self
        key = self.cls.construct(id=last_doc_id)._key
        snapshot = db.conn.document(key).get()
        self.firebase_query = self.firebase_query.start_after(snapshot)
        return self

    def start_after(self, fields):
        self.firebase_query = self.firebase_query.start_after(fields)
        return self

    def end_at(self, fields):
        self.firebase_query = self.firebase_query.end_at(fields)
        return self

    def end_before(self, fields):
        self.firebase_query = self.firebase_query.end_before(fields)
        return self

    def limit(self, limit):
        print("calling limit", limit)
        if not limit or limit == -1:
            return self
        self.firebase_query = self.firebase_query.limit(limit)
        return self

    def add_q(self, q: Q = None):
        if not q:
            return self
        self.limit(q.limit)
        self.start_after_id(q.start_after_id)
        return self

    def get_path_of_query(self):
        if getattr(self.firebase_query, "_path", None):
            return "/".join(self.firebase_query._path)
        if getattr(self.firebase_query, "_parent", None):
            return "/".join(self.firebase_query._parent._path)
        else:
            return self.collection_path

    """Querying and creating of the models."""

    def make_safe_span_label(self, query_type):
        return f"{query_type}-{self.get_path_of_query()}-{str(self.query_inputs)}"

    @magic_async
    def stream(self, validate_db_data=True, **kwargs):
        """Currently defaults to validating the db data, but you can turn this off with the validate_db_data param"""
        with safe_span(self.make_safe_span_label("stream")):
            docs = list(self.firebase_query.stream(**kwargs))

        constructor = self.cls if validate_db_data else self.cls.construct

        return [
            constructor(
                from_db=True, _doc=doc, **doc.to_dict(), _key=doc.reference.path
            )
            for doc in docs
        ]

    @magic_async
    def get(
        self,
        id=None,
        create=False,
        validate_db_data=True,
        skip_redis=False,
        skip_redis_get=False,
        skip_redis_set=False,
        **kwargs,
    ):
        if skip_redis:
            skip_redis_get = True
            skip_redis_set = True
        if id:
            self.firebase_query = self.firebase_query.document(id)

        d = None
        redis_miss = False

        _key = self.get_path_of_query()
        _doc = None

        if self.redis and not skip_redis_get:
            with safe_span(self.make_safe_span_label("get-redis")):
                raw = self.redis.get(_key)
                if not raw:
                    redis_miss = True
                    print(f"Redis miss for {self.get_path_of_query()}.")
                else:
                    d = ujson.loads(raw)
        if not d:
            with safe_span(self.make_safe_span_label("get")):
                doc = self.firebase_query.get(**kwargs)
            if type(doc) != DocumentSnapshot:
                return None
            d = doc.to_dict()
            _doc = doc
            _key = doc.reference.path

        if not d:
            return None if not create else self.cls(id=id)

        constructor = self.cls if validate_db_data else self.cls.construct

        model = constructor(
            from_db=True,
            _doc=_doc,
            **d,
            _key=_key,
        )

        if redis_miss and not skip_redis_set:
            if self.redis_ttl_secs:
                with safe_span(self.make_safe_span_label("setex-redis")):
                    self.redis.setex(
                        name=self.get_path_of_query(),
                        time=self.redis_ttl_secs,
                        value=model.json(),
                    )
            else:
                with safe_span(self.make_safe_span_label("set-redis")):
                    self.redis.set(name=self.get_path_of_query(), value=model.json())

        return model

    @magic_async
    def get(
        self,
        id=None,
        create=False,
        validate_db_data=True,
        skip_redis=False,
        skip_redis_get=False,
        skip_redis_set=False,
        **kwargs,
    ):
        if skip_redis:
            skip_redis_get = True
            skip_redis_set = True
        if id:
            self.firebase_query = self.firebase_query.document(id)

        d = None
        redis_miss = False

        _key = self.get_path_of_query()
        _doc = None

        if self.redis and not skip_redis_get:
            with safe_span(self.make_safe_span_label("get-redis")):
                raw = self.redis.get(_key)
                if not raw:
                    redis_miss = True
                    print(f"Redis miss for {self.get_path_of_query()}.")
                else:
                    d = ujson.loads(raw)
        if not d:
            with safe_span(self.make_safe_span_label("get")):
                doc = self.firebase_query.get(**kwargs)
            if type(doc) != DocumentSnapshot:
                return None
            d = doc.to_dict()
            _doc = doc
            _key = doc.reference.path

        if not d:
            return None if not create else self.cls(id=id)

        constructor = self.cls if validate_db_data else self.cls.construct

        model = constructor(
            from_db=True,
            _doc=_doc,
            **d,
            _key=_key,
        )

        if redis_miss and not skip_redis_set:
            if self.redis_ttl_secs:
                with safe_span(self.make_safe_span_label("setex-redis")):
                    self.redis.setex(
                        name=self.get_path_of_query(),
                        time=self.redis_ttl_secs,
                        value=model.json(),
                    )
            else:
                with safe_span(self.make_safe_span_label("set-redis")):
                    self.redis.set(name=self.get_path_of_query(), value=model.json())

        return model

    @magic_async
    def get_all(
        self,
        ids,
        validate_db_data=True,
        skip_redis=False,
        skip_redis_get=False,
        skip_redis_set=False,
        **kwargs,
    ):
        if skip_redis:
            skip_redis_get = True
            skip_redis_set = True

        doc_refs = [db.conn.collection(self.collection_path).document(id) for id in ids]
        _keys = [doc_ref._path for doc_ref in doc_refs]

        docs = None
        redis_miss = False
        d_arr = None
        if self.redis and not skip_redis_get:
            with safe_span(self.make_safe_span_label("get_all-redis")):
                raw_arr = self.redis.mget(_keys)
                if None in raw_arr:
                    redis_miss = True
                    print(f"Redis miss for {self.get_path_of_query()}.")
                else:
                    d_arr = [ujson.loads(raw) for raw in raw_arr]
        if not d_arr:
            with safe_span(self.make_safe_span_label("get_all")):
                docs = list(db.conn.get_all(doc_refs, **kwargs))

        constructor = self.cls if validate_db_data else self.cls.construct

        models = []
        if d_arr:
            for key, d in zip(_keys, d_arr):
                models.append(constructor(from_db=True, _key=key, **d))
        else:
            for doc in docs:
                models.append(
                    constructor(
                        from_db=True, _key=doc.reference.path, _doc=doc, **doc.to_dict()
                    )
                )

        """
        models = [
            constructor(
                from_db=True,
                _doc=doc,
                **doc.to_dict(),
                _key=doc.reference.path,
            )
            for doc in docs
        ]
        """
        sorted_models = self.sort_models_by_id_order(models, ids)
        if redis_miss and not skip_redis_set:
            model_d = {model.key: model.json() for model in sorted_models}
            if self.redis_ttl_secs:
                with safe_span(
                    self.make_safe_span_label(f"msetex-redis-{len(model_d)}")
                ):
                    self.redis.setex(
                        name=self.get_path_of_query(),
                        time=self.redis_ttl_secs,
                        value=model.json(),
                    )
            else:
                with safe_span(self.make_safe_span_label("set-redis")):
                    self.redis.set(name=self.get_path_of_query(), value=model.json())

            self.redis.mset(model_d)

    @staticmethod
    def sort_models_by_id_order(models, ids):
        """Maintain order of return array"""
        sorted_models = [None] * len(ids)
        sort_key = {id: index for index, id in enumerate(ids)}
        for model in models:
            sorted_models[sort_key[model.id]] = model
        return sorted_models

    """Subcollections"""

    def collections(self):
        return self.firebase_query.collections()

    def collections_names(self):
        return [coll.id for coll in self.firebase_query.collections()]



import time
import typing
from datetime import datetime
import pprint
from magicdb.Models import MagicModel, DateModel
import magicdb
import firebase_admin

from firebase_admin import firestore, credentials

# creds = credentials.Certificate("/Users/jeremyberman/Downloads/my-service-account.json")

# firebase_admin.initialize_app(creds)

# db = firestore.client()

import os

os.environ[
    "GOOGLE_APPLICATION_CREDENTIALS"
] = "/Users/jeremyberman/Downloads/my-service-account.json"


class TestModel(MagicModel):
    # ref = db.collection("new_artists_9_16").document("1")
    name = "frankR"
    old_beatgig_data: typing.Dict = None
    spotify_api_data: typing.Dict = None

    class Meta:
        collection_name = "new_artists_9_16"
        # magic_fields_to_exclude = {"ref"}


class Change(MagicModel):
    name: str
    created_at: datetime

    class Meta:
        collection_name = "changes"


"""
# test_model = TestModel(id="1")
# test_model2 = TestModel(id="2", parent=test_model)
# print("22222", test_model2)
# print(test_model.collection_name)

print("ddd", test_model.dict())
print(test_model.json())
# print("ssss", test_model.schema())
print(test_model.ref.path)
# print("ssss", test_model2.schema())
print("jjjj", test_model.schema_json())
"""
# pprint.pprint(test_model.ref.__dict__)

# ref = db.collection("new_artists_9_16").document("1")
# print(type(ref) == magicdb.DocumentReference)
# print("ref", ref)

# models = TestModel.collection.limit(10).stream(validate_db_data=False)
# print(models)
# print(str(models[0]))

# m = TestModel.collection.get_all(ids=["123456", "1"])
# print("m", m)


def test_harber():
    harber_test = TestModel.collection.get(id="123456", validate_db_data=False)
    print("hhh", harber_test)
    print(type(harber_test))
    changes = (
        Change.collection.parent(harber_test)
        .order_by("created_at", "desc")
        .limit(2)
        .stream()
    )

    print("changes", changes)
    print(list(changes))

    print(harber_test.get_subcollections()[0].__dict__)

    start = time.time()
    new_test: TestModel = TestModel(**harber_test.dict())
    print("took", time.time() - start)
    print(new_test)

    q = Change.collection.parent(harber_test).order_by("created_at", "desc").limit(2)

    res = Change.stream_queries([q])
    print("RESSSS", res)


# test_harber()


class Teacher(MagicModel):
    name: str
    scuba: str = "snacks"
    hurr: str = None

    class Meta:
        collection_name = "teachers_testing"


def test_teacher():
    t = Teacher(name="sally")
    # t.save()
    print(t.__dict__)
    t.save()

    t.name = "henty"
    print("ttt", t)
    t.update()
    print("tttt", t)
    t.scuba = "rob"
    t.update()
    t.hurr = "['hi']"
    print("hu", t)
    t.update()
    t.save()

    h = Teacher(name="h")
    h.set_id("23")
    h.save()
    t.set_parent(h)
    print("tttttt", t)
    t.print_all()
    t.set_id("777")
    t.update(create=True)
    t.hurr = "uuuu"

    t.update()


class Dog(DateModel):
    name: str = "jon"
    has_bone: bool = False


def test_dog():
    dog1 = Dog()
    print(dog1)
    dog1.name = "tobu"
    print(dog1)


# test_dog()



import magicdb
from magicdb import DELETE_FIELD


def make_update_obj_rec(original, new, current_path, update_d):
    field_path = magicdb.db.conn.field_path

    # first get all the deletions of fields
    del_fields = original.keys() - new.keys()
    for field in del_fields:

        update_d[".".join([*current_path, field_path(field)])] = DELETE_FIELD

    # now get all the object and value changes
    for field, new_value in new.items():
        if field not in original:
            update_d[".".join([*current_path, field_path(field)])] = new_value
            continue
        original_value = original[field]
        if new_value == original_value:
            continue
        # this is the case where they are different...
        if type(new_value) != type(original_value) or not type(new_value) == dict:
            update_d[".".join([*current_path, field_path(field)])] = new_value
            continue
        # now you know this is a dict and it's different
        make_update_obj_rec(
            original_value, new_value, [*current_path, field_path(field)], update_d
        )


def make_update_obj(original, new):
    update_d = {}
    make_update_obj_rec(original=original, new=new, current_path=[], update_d=update_d)
    return update_d



import concurrent.futures

import asyncio
from functools import wraps

_DEFAULT_POOL = concurrent.futures.ThreadPoolExecutor(max_workers=20)
# _DEFAULT_POOL._max_workers = max(_DEFAULT_POOL._max_workers, 20)


def magic_async(f, executor=None):
    """Makes a sync function async by returning asyncio futures that can be awaited.
    Function takes in a sync param and if False, will return an awaitable promise"""

    @wraps(f)
    def wrap(*args, sync=True, future=False, **kwargs):
        if sync and not future:
            return f(*args, **kwargs)
        future_res = (executor or _DEFAULT_POOL).submit(f, *args, **kwargs)
        if future:
            return future_res
        return asyncio.wrap_future(future_res)

    return wrap


def get_futures(futures):
    return [future.result() for future in futures]


def threadpool_asyncio(f, executor=None):
    """Makes a sync function async by returning asyncio futures that can be awaited."""

    @wraps(f)
    def wrap(*args, **kwargs):
        future = (executor or _DEFAULT_POOL).submit(f, *args, **kwargs)
        return asyncio.wrap_future(future)

    return wrap


def threadpool(f, executor=None):
    @wraps(f)
    def wrap(*args, **kwargs):
        future = (executor or _DEFAULT_POOL).submit(f, *args, **kwargs)
        return future

    return wrap


def promise(f, executor=None):
    """Makes async function a task -- to be executed concurrently."""

    @wraps(f)
    def wrap(*args, **kwargs):
        return asyncio.create_task(f(*args, **kwargs))

    return wrap



from magicdb.Models.MagicModel import QueryAndBaseMetaClasses


class RemoveFieldsMeta(QueryAndBaseMetaClasses, type):
    @staticmethod
    def remove_fields(cls):
        if "fields_to_remove" not in cls.__fields__:
            return
        for field in cls.__fields__["fields_to_remove"].default:
            if field in cls.__fields__:
                del cls.__fields__[field]
        del cls.__fields__["fields_to_remove"]

    def __new__(mcs, *args, **kwargs):
        x = super().__new__(mcs, *args, **kwargs)
        mcs.remove_fields(x)
        return x



from __future__ import annotations
import threading
from copy import deepcopy
import json
from typing import List, Dict, Any
import os
import concurrent.futures
from glom import glom
from datetime import datetime
import copy
import time
import pprint


import magicdb
from pydantic import BaseModel, Field
from pydantic.main import ModelMetaclass

from magicdb.Models import model_helpers
from magicdb.Queries import Query
from magicdb.utils.Serverless.span import safe_span

from magicdb.utils.updating_objects import make_update_obj

from magicdb.utils.async_helpers import threadpool_asyncio, threadpool, magic_async


from slugify import slugify as base_slugify

slugify_kwargs = {"separator": "", "replacements": [["&", "and"]]}

MAGIC_FIELDS = ["id", "key", "ref", "parent", "kwargs_from_db", "doc"]

# ref will break this cause it is not jsonable
MAGIC_FIELDS_TO_EXCLUDE = set(MAGIC_FIELDS) - {"id"}

FIELDS_TO_EXCLUDE_FOR_DB = set(MAGIC_FIELDS)


class DatabaseError(Exception):
    def __init__(self, key):
        self.message = (
            f"There is no document with key {key} to update. Add update.(create=True) to save the document"
            f" if it does not exist. Otherwise, you can save the document: save()."
        )


class QueryMeta(type):
    """https://stackoverflow.com/questions/128573/using-property-on-classmethods"""

    @property
    def collection(cls) -> Query:
        return Query(cls)

    @property
    def collection_group(cls) -> Query:
        return Query(cls).collection_group()


class QueryAndBaseMetaClasses(ModelMetaclass, QueryMeta):
    pass


class MagicModel(BaseModel, metaclass=QueryAndBaseMetaClasses):
    """
    When this gets inited, if given an id or key, assign based on that.
    Otherwise, assign them based on what Firestore gives it
    """

    id: str = None
    key: str = None
    # ref: magicdb.DocumentReference = None
    # doc: magicdb.DocumentSnapshot = None
    ref: Any = None
    doc: Any = None
    parent: MagicModel = None
    kwargs_from_db: dict = None

    __call__ = ...  # to satisfy Query python linter

    def __init__(self, from_db: bool = False, **kwargs):
        """Feed in all magic fields as kwargs"""
        # TODO test how slow this deep copy is
        kwargs_from_db = {} if not from_db else deepcopy(kwargs)
        model_helpers.make_magic_fields_from_kwargs(
            kwargs=kwargs, collection_name=self.collection_name
        )
        super().__init__(**kwargs)
        self.kwargs_from_db = kwargs_from_db

    @classmethod
    def construct(cls, *args, from_db: bool = False, **kwargs):
        kwargs_from_db = {} if not from_db else deepcopy(kwargs)
        model_helpers.make_magic_fields_from_kwargs(
            kwargs=kwargs, collection_name=cls.get_collection_name()
        )
        new_obj = super().construct(*args, **kwargs)
        new_obj.kwargs_from_db = kwargs_from_db
        return new_obj

    """GETTING AND SETTING FIELDS"""

    def set_id(self, id: str):
        kwargs = {"id": id}
        model_helpers.make_magic_fields_from_id_parent_or_nothing(
            kwargs=kwargs, collection_name=self.collection_name
        )
        self.__dict__.update(kwargs)

    def set_key(self, key: str):
        kwargs = {"key": key}
        model_helpers.make_magic_fields_from_key(kwargs=kwargs)
        self.__dict__.update(kwargs)

    def set_ref(self, ref: magicdb.DocumentReference):
        kwargs = {"ref": ref}
        model_helpers.make_magic_fields_from_ref(kwargs=kwargs)
        self.__dict__.update(kwargs)

    def set_parent(self, parent: MagicModel):
        kwargs = {"parent": parent}
        model_helpers.make_magic_fields_from_id_parent_or_nothing(
            kwargs=kwargs, collection_name=self.collection_name
        )
        self.__dict__.update(kwargs)

    """OVERRIDING PYDANTIC"""

    @classmethod
    def get_fields_to_exclude(cls):
        magic_fields_to_exclude = getattr(
            cls.Meta, "magic_fields_to_exclude", MAGIC_FIELDS_TO_EXCLUDE
        )
        given_fields_to_exclude = getattr(cls.Meta, "exclude", set())
        return magic_fields_to_exclude | given_fields_to_exclude

    def dict(self, exclude_magic_fields=True, *args, **kwargs):
        to_exclude = set() if not exclude_magic_fields else self.get_fields_to_exclude()
        # join w other fields passed in
        kwargs["exclude"] = kwargs.get("exclude") or set() | to_exclude
        return super().dict(*args, **kwargs)

    @classmethod
    def schema(cls, *args, **kwargs):
        """Temporarily take out the excluded fields to get the schema, then put them back."""

        original_fields = cls.__fields__.copy()

        for magic_field in cls.get_fields_to_exclude():
            if magic_field in cls.__fields__:
                del cls.__fields__[magic_field]

        schema_d = super().schema(*args, **kwargs)

        # cannot set __fields__ directly so will remove all items then update the d w the original fields
        for key in list(cls.__fields__.keys()):
            del cls.__fields__[key]

        cls.__fields__.update(original_fields)
        return schema_d

    """META CLASS FUNCTIONS"""

    @property
    def collection_name(self) -> str:
        return self.get_collection_name()

    @classmethod
    def make_default_collection_name(cls) -> str:
        return cls.__name__.lower()

    @classmethod
    def get_collection_name(cls) -> str:
        return getattr(cls.Meta, "collection_name", cls.make_default_collection_name())

    """PRINTING AND RETURNING"""

    def __repr__(self, *args, **kwargs):
        return f"{self.__class__.__name__}({self.__repr_str__(', ')})"

    def __str__(self, *args, **kwargs):
        return f"{self.__repr_str__(' ')}"

    def __repr_str__(self, join_str=", ", fields_to_exclude: set = None):
        fields_to_exclude: set = (
            fields_to_exclude
            if fields_to_exclude is not None
            else self.get_fields_to_exclude()
        )
        key_values: List[str] = []
        for field in self.__dict__:
            if field not in fields_to_exclude:
                key_values.append(f"{repr(field)}={repr(getattr(self, field, None))}")
        return join_str.join(key_values)

    def print_all(self):
        result: str = self.__repr_str__(fields_to_exclude=set())
        print("PRINT_ALL", result)
        return result

    """COMPARISONS"""

    def equal(self, model: BaseModel, path_to_compare=None):
        """The difference between objects will be firestore adds nanoseconds to the object. So parse those out"""
        replace_text = "+00:00"
        self_json = self.json().replace(replace_text, "")
        other_json = model.json().replace(replace_text, "")
        if path_to_compare:
            self_json = json.dumps(
                glom(json.loads(self_json), ".".join(path_to_compare))
            )
            other_json = json.dumps(
                glom(json.loads(other_json), ".".join(path_to_compare))
            )
        return self_json == other_json

    def been_updated(self, path_to_compare=None):
        from_db = self.__class__.get_with_error(self.id)
        return not self.equal(from_db, path_to_compare=path_to_compare)

    class Meta:
        """Init the meta class so you can use it and know it is there"""

        """Looks like having Meta in the TestModel actually overrites this one, not inherits it"""
        ...

    class Serverless:
        """Fills this with the latest context if it exists"""

        context = None

    class Config:
        anystr_strip_whitespace: bool = True

        arbitrary_types_allowed: bool = True

        @staticmethod
        def schema_extra(schema: Dict[str, Any], model: MagicModel) -> None:
            keys = list(schema.get("properties", {}).keys())
            to_exclude = model.get_fields_to_exclude()
            for key in keys:
                if key in to_exclude:
                    schema.get("properties", {}).pop(key)

        json_encoders = {magicdb.DocumentReference: lambda doc_ref: doc_ref.path}

    """ADDING TO FIRESTORE"""

    @staticmethod
    def remove_magic_fields(d):
        for magic_field in MAGIC_FIELDS:
            if magic_field in d:
                del d[magic_field]
        return d

    def save(self, batch=None, transaction=None, merge=False, ignore_fields=False):
        """Will create a new obj_to_save and save it so that all of the validation happens properly on a new obj."""
        batch_or_transaction = batch or transaction
        obj_to_save = (
            self
            if ignore_fields
            else self.__class__(**self.dict(exclude_magic_fields=False))
        )
        new_d = obj_to_save.dict()
        self.remove_magic_fields(new_d)

        with safe_span(f"save-{self.key}", use=(batch_or_transaction is None)):
            obj_to_save.ref.set(
                new_d, merge=merge
            ) if not batch_or_transaction else batch_or_transaction.set(
                obj_to_save.ref, new_d, merge=merge
            )
        if not merge:
            obj_to_save.kwargs_from_db = copy.deepcopy(new_d)

        # update self just in case
        self.__dict__.update(obj_to_save.__dict__)
        return obj_to_save

    def update(
        self,
        batch=None,
        transaction=None,
        create=False,
        ignore_fields=False,
        print_update_d=False,
        only_print_update_d=False,
    ):
        batch_or_transaction = batch or transaction
        # TODO this takes a long time... find out why and fix! Do you need to make new class?
        obj_to_update = (
            self
            if ignore_fields
            else self.__class__(**self.dict(exclude_magic_fields=False))
        )
        new_d = obj_to_update.dict()
        kwargs_d = self.kwargs_from_db

        self.remove_magic_fields(new_d)
        self.remove_magic_fields(kwargs_d)

        self.get_fields_to_exclude()
        update_d = (
            new_d if not kwargs_d else make_update_obj(original=kwargs_d, new=new_d)
        )
        if print_update_d or only_print_update_d:
            pprint.pprint({f"update_d_for_{self.id}": update_d})
        if only_print_update_d:
            return
        try:
            if update_d != {}:
                with safe_span(
                    f"update-{self.key}", use=(batch_or_transaction is None)
                ):
                    obj_to_update.ref.update(
                        update_d
                    ) if not batch_or_transaction else batch_or_transaction.update(
                        obj_to_update.ref, update_d
                    )
            else:
                print(f"update_d was empty so did not update for {self.id}.")
            self.__dict__.update(obj_to_update.__dict__)
            return obj_to_update
        except Exception as e:
            if hasattr(e, "message") and "no document to update" in e.message.lower():
                if create:
                    return obj_to_update.save(batch=batch, transaction=transaction)
                else:
                    db_error = DatabaseError(obj_to_update.key)
                    raise DatabaseError(db_error.message)
            raise e

    def delete(self, batch=None, transaction=None):
        batch_or_transaction = batch or transaction
        with safe_span(f"delete-{self.key}", use=(batch_or_transaction is None)):
            return (
                self.ref.delete()
                if not batch_or_transaction
                else batch_or_transaction.delete(self.ref)
            )

    """Save history to DB"""

    def save_history(self, sync: bool = True, user_id: str = None, **kwargs):
        """Saves current object to a changes sub collection"""
        history_key = f"{self.key}/history/{str(int(time.time()))}"

        class History(self.__class__):
            changed_at: datetime = Field(default_factory=datetime.utcnow)
            changed_by: str = user_id

        history_obj = History(**self.dict(), key=history_key)
        return history_obj.save(**kwargs) if sync else history_obj.save_async(**kwargs)

    def save_deleted(self, sync: bool = True, user_id: str = None, **kwargs):
        deleted_key = f"deleted_{self.collection_name}/{self.id}"

        class Deleted(self.__class__):
            deleted_at: datetime = Field(default_factory=datetime.utcnow)
            deleted_by: str = user_id

        deleted_obj = Deleted(**self.dict(), key=deleted_key)
        return deleted_obj.save(**kwargs) if sync else deleted_obj.save_async(**kwargs)

    """Writing to DB async"""

    def save_async(self, *args, **kwargs) -> threading.Thread:
        t = threading.Thread(target=self.save, args=(*args,), kwargs={**kwargs})
        t.start()
        return t

    def update_async(self, *args, **kwargs) -> threading.Thread:
        t = threading.Thread(target=self.update, args=(*args,), kwargs={**kwargs})
        t.start()
        return t

    def delete_async(self, *args, **kwargs) -> threading.Thread:
        t = threading.Thread(target=self.delete, args=(*args,), kwargs={**kwargs})
        t.start()
        return t

    """QUERYING AND COLLECTIONS"""

    def exists(self):
        return self.__class__.collection.get(self.id) is not None

    def get_subcollections(self):
        return list(self.__class__.collection.document(self.id).collections())

    @classmethod
    def get_ref(cls, id) -> magicdb.DocumentReference:
        temp_cls = cls.construct(id=id)
        return temp_cls.ref

    @classmethod
    def get_default_exception(cls):
        return getattr(cls.Meta, "default_exception", Exception)

    @classmethod
    @magic_async
    def get_with_error(cls, id, error=None, **kwargs):
        if not error:
            error = cls.get_default_exception()
        val = cls.collection.get(id, **kwargs)
        if not val:
            raise error(f"{cls.__name__} with id {id} does not exist.")
        return val

    @classmethod
    @threadpool_asyncio
    def get_with_error_async(cls, id, error=None, **kwargs):
        if not error:
            error = cls.get_default_exception()
        val = cls.collection.get(id, **kwargs)
        if not val:
            raise error(f"{cls.__name__} with id {id} does not exist.")
        return val

    @classmethod
    @threadpool_asyncio
    def get_async(cls, id, **kwargs):
        return cls.collection.get(id, **kwargs)

    @classmethod
    @threadpool
    def get_with_error_future(cls, id, error=None, **kwargs):
        if not error:
            error = cls.get_default_exception()
        val = cls.collection.get(id, **kwargs)
        if not val:
            raise error(f"{cls.__name__} with id {id} does not exist.")
        return val

    """GETTING SUBCLASSES"""

    @classmethod
    def get_subclasses(cls):
        all_subs = []
        for sub in cls.__subclasses__():
            all_subs.append(sub)
            all_subs += sub.get_subclasses()
        return list(set(all_subs))

    @staticmethod
    def get_all_subclasses_of_model():
        all_subs = []
        for sub in list(MagicModel.__subclasses__()):
            all_subs.append(sub)
            all_subs += sub.get_subclasses()
        return list(set(all_subs))

    @staticmethod
    def stream_queries(queries: List[magicdb.Query]):
        with concurrent.futures.ThreadPoolExecutor(max_workers=20) as executor:
            # set this to 10 because lambda only has 2 cores so would only have 6 threads for max_workers
            # executor._max_workers = max(executor._max_workers, 20)
            futures = [
                executor.submit(
                    query.stream,
                )
                for query in queries
            ]
            return [f.result() for f in futures]

    """SLUG STUFF"""

    @classmethod
    def slugify(cls, slug):
        return base_slugify(slug, **slugify_kwargs)

    @classmethod
    def create_slug_static(cls, value_to_slugify: str):
        slug_base = cls.slugify(value_to_slugify.lower().replace(" ", ""))
        number = 1
        slug = slug_base
        while cls.get_by_slug(slug):
            slug = cls.slugify(f"{slug_base}{number}")
            number += 1
        return slug

    def create_slug(self, field_to_slugify: str = "name"):
        base_val = getattr(self, field_to_slugify, "")
        slug_base = self.__class__.slugify(base_val.lower().replace(" ", ""))
        number = 1
        slug = slug_base
        while not self.unique_slug(slug):
            slug = self.__class__.slugify(f"{slug_base}{number}")
            number += 1
        self.slug = slug
        return self.slug

    def unique_slug(self, slug):
        models = self.__class__.collection.where("slug", "==", slug).stream()
        for model in models:
            if model.id != self.id:
                return False
        return True

    @classmethod
    def get_by_slug(cls, slug):
        models = cls.collection.where("slug", "==", slug).stream()
        return None if not models else models[0]

    @classmethod
    def get_by_slug_with_error(cls, slug, error=None):
        if not error:
            error = cls.get_default_exception()
        models = cls.collection.where("slug", "==", slug).stream()
        if not models:
            raise error(f"{cls.__name__} with slug {slug} does not exist.")
        return models[0]

    @classmethod
    @magic_async
    def get_by_field(cls, field_name: str, value):
        models = cls.collection.where(
            magicdb.db.conn.field_path(field_name), "==", value
        ).stream()
        return None if not models else models[0]

    @classmethod
    @threadpool_asyncio
    def get_by_field_async(cls, field_name: str, value):
        return cls.get_by_field(field_name=field_name, value=value)

    @classmethod
    def get_by_field_with_error(cls, field_name: str, value, error=None):
        if not error:
            error = cls.get_default_exception()
        model = cls.get_by_field(field_name=field_name, value=value)
        if not model:
            raise error(f"{cls.__name__} with {field_name} {value} does not exist.")
        return model

    @classmethod
    @threadpool_asyncio
    def get_by_field_with_error_async(cls, field_name: str, value, error=None):
        return cls.get_by_field_with_error(
            field_name=field_name, value=value, error=error
        )


MagicModel.update_forward_refs()



from pydantic import BaseModel, validator


class FrontendParserModel(BaseModel):
    """
    Model turns all optional empty strings to None, removes all
    empty string or None elements from an array, and strips whitespace.
    """

    @staticmethod
    def empty_str(v):
        return type(v) == str and v.strip() == ""

    @validator("*", pre=True)
    def empty_str_to_none(cls, v, field):
        if (
            cls.empty_str(v)
            and not field.required
            and getattr(field, "default", True) is None
        ):
            v = None
        return v

    @validator("*", pre=True)
    def remove_empty_string_and_nones_from_list(cls, v):
        if type(v) != list:
            return v
        new_v = []
        """If list -- only include valid values -- and strip strs"""
        for e in v:
            if e is None or cls.empty_str(e):
                continue
            if type(e) == str:
                e = e.strip()
            new_v.append(e)
        return new_v

    class Config:
        anystr_strip_whitespace = True



from __future__ import annotations
import typing as T

import redis

from .QuerySpeed import QuerySpeed

from magicdb.utils.Serverless.span import safe_span
from magicdb.utils.async_helpers import magic_async


class QueryRedis(QuerySpeed):
    ...

    def get_d_from_db(self, **kwargs) -> dict:
        # first try redis, then do firestore if miss
        key = self.get_path_of_query()

        with safe_span(self.make_safe_span_label("get-redis")):
            if d := self.r.get(key):
                print("Redis hit!")
                return d

        print("Redis miss!")
        # there was a cache miss
        model = super().get_d_from_db(**kwargs)
        if model:
            with safe_span(self.make_safe_span_label("set-redis")):
                self.r.set(key)

    def get(self, id=None, *args, **kwargs):
        # if cache miss
        if id:
            self.firebase_query = self.firebase_query.document(id)
        key = self.get_path_of_query()

        if d := self.r.get(key):
            ...

        model = super().get(*args, **kwargs)
        self.r.set(key, model.json())
        return model

    @property
    def r(self) -> T.Optional[redis.Redis]:
        return self.cls.get_redis()

    @property
    def r_ttl_secs(self) -> T.Optional[int]:
        return self.cls.get_redis_ttl_secs()



__version__ = "0.1.0"

# first import all of the necessary firebase stuff
from firebase_admin import firestore, credentials, auth, storage


ArrayUnion = firestore.firestore.ArrayUnion
ArrayRemove = firestore.firestore.ArrayRemove
DESCENDING = firestore.firestore.Query.DESCENDING
ASCENDING = firestore.firestore.Query.ASCENDING
Increment = firestore.firestore.Increment
SERVER_TIMESTAMP = firestore.firestore.SERVER_TIMESTAMP
DELETE_FIELD = firestore.firestore.DELETE_FIELD

WriteBatch = firestore.firestore.WriteBatch

DocumentReference = firestore.firestore.DocumentReference
CollectionReference = firestore.firestore.CollectionReference

DocumentSnapshot = firestore.firestore.DocumentSnapshot

Query = firestore.firestore.Query

# give access to connect right from magicdb
from magicdb.database import connect, db

from magicdb.FirestoreWrappers.BatchWrapper import BatchWrapper


# to connect to the raw db, do db.conn
def batch(*args, **kwargs):
    # return db.conn.batch(*args, **kwargs)
    return BatchWrapper(*args, **kwargs)


def transaction(*args, **kwargs):
    return db.conn.transaction(*args, **kwargs)


from magicdb.Models import (
    MagicCache,
    MagicModel,
    DateModel,
    FrontendParserModel,
    RemoveFieldsMeta,
    RemoveFieldsMetaSpeed,
    MagicModelSpeed,
    DateModelSpeed,
)

from magicdb.Queries import Q

from magicdb.utils.async_helpers import (
    magic_async,
    threadpool_asyncio,
    threadpool,
    promise,
    get_futures,
)



import os
from functools import wraps
from time import time
from magicdb.Timing.main import function_times

TIMEIT = os.getenv("TIMEIT")


def timing(f):
    @wraps(f)
    def wrap(*args, **kw):
        if not TIMEIT:
            return f(*args, **kw)
        start = time()
        result = f(*args, **kw)
        end = time()
        time_took = end - start
        function_times[f.__name__] = function_times.get(f.__name__, 0) + time_took
        print("function_times", function_times)
        return result

    return wrap



import firebase_admin
from firebase_admin import credentials, firestore


class Database:
    def __init__(self):
        self._conn = None

    def connect(self, creds=None, from_file=None, firestore_instance=None):
        if not creds and not from_file and not firestore_instance:
            raise Exception(
                "Credentials, service account json file path, or firestore_instance required to connect with firestore."
            )
        if from_file:
            creds = credentials.Certificate(from_file)
        if creds:
            try:
                firebase_admin.initialize_app(creds)
            except Exception as e:
                if (
                    "The default Firebase app already exists" in str(e)
                    and not firestore_instance
                ):
                    raise Exception(
                        """
                        If you want to connect to Firestore from_file, make sure magicdb.connect(from_file=<YOUR FILE>)
                        comes directly after importing magicdb for the first time.
                        """
                    )
        if firestore_instance:
            self._conn = firestore_instance
        else:
            self._conn = firestore.client()

    @property
    def conn(self):
        if self._conn is None:
            firebase_admin.initialize_app()
            self._conn = firestore.client()
        return self._conn



import typing as T
from typing import Union, List
import magicdb
from magicdb import db, ASCENDING, DESCENDING, DocumentSnapshot
from magicdb.Queries.Q import Q
from magicdb.utils.Serverless.span import safe_span
from magicdb.utils.async_helpers import magic_async

import redis

import ujson

import time


class QuerySpeed:
    def __init__(self, cls):
        from magicdb.Models.MagicModelSpeed import MagicModelSpeed

        self.cls: MagicModelSpeed = cls
        self.collection_path = self.cls.get_collection_name()
        self.firebase_query = self.create_query()
        self.query_inputs: List[str] = []

    @property
    def key(self):
        return self.get_path_of_query()

    def create_query(self):
        self.firebase_query = db.conn.collection(self.collection_path)
        return self.firebase_query

    def parent(self, parent_or_path):
        from magicdb.Models.MagicModelSpeed import MagicModelSpeed

        parent_path = (
            parent_or_path
            if not issubclass(type(parent_or_path), MagicModelSpeed)
            else parent_or_path._key
        )

        self.collection_path = parent_path + "/" + self.collection_path
        self.create_query()
        return self

    def document(self, document_path):
        self.firebase_query = self.firebase_query.document(document_path)
        return self

    def collection_group(self):
        self.firebase_query = db.conn.collection_group(self.cls.get_collection_name())
        return self

    def collection(self, collection_path):
        self.firebase_query = self.firebase_query.collection(collection_path)
        return self

    """Wrapping Firestore abilities."""

    def where(self, field, action, value):
        self.firebase_query = self.firebase_query.where(field, action, value)
        self.query_inputs.append(f"{field}, {action}, {value}")
        return self

    def order_by(self, field, direction=None, **kwargs):
        if direction:
            kwargs["direction"] = direction
        if kwargs.get("direction", None) == "asc":
            kwargs["direction"] = ASCENDING
        if kwargs.get("direction", None) == "desc":
            kwargs["direction"] = DESCENDING
        self.firebase_query = self.firebase_query.order_by(field, **kwargs)
        return self

    def start_at(self, fields):
        self.firebase_query = self.firebase_query.start_at(fields)
        return self

    def start_after_id(self, last_doc_id: str = None):
        if not last_doc_id:
            return self
        key = self.cls.construct(id=last_doc_id)._key
        snapshot = db.conn.document(key).get()
        self.firebase_query = self.firebase_query.start_after(snapshot)
        return self

    def start_after(self, fields):
        self.firebase_query = self.firebase_query.start_after(fields)
        return self

    def end_at(self, fields):
        self.firebase_query = self.firebase_query.end_at(fields)
        return self

    def end_before(self, fields):
        self.firebase_query = self.firebase_query.end_before(fields)
        return self

    def limit(self, limit):
        print("calling limit", limit)
        if not limit or limit == -1:
            return self
        self.firebase_query = self.firebase_query.limit(limit)
        return self

    def add_q(self, q: Q = None):
        if not q:
            return self
        self.limit(q.limit)
        self.start_after_id(q.start_after_id)
        return self

    def get_path_of_query_old(self):
        if getattr(self.firebase_query, "_path", None):
            return "/".join(self.firebase_query._path)
        if getattr(self.firebase_query, "_parent", None):
            return "/".join(self.firebase_query._parent._path)
        else:
            return self.collection_path

    def get_path_of_query(self):
        if path := getattr(self.firebase_query, "path", None):
            return path
        if parent := getattr(self.firebase_query, "_parent", None):
            return parent.path
        else:
            return self.collection_path

    """Querying and creating of the models."""

    def make_safe_span_label(self, query_type):
        return f"{query_type}-{self.get_path_of_query()}-{str(self.query_inputs)}"

    @magic_async
    def stream(self, validate_db_data=True, **kwargs):
        """Currently defaults to validating the db data, but you can turn this off with the validate_db_data param"""
        with safe_span(self.make_safe_span_label("stream")):
            docs = list(self.firebase_query.stream(**kwargs))

        constructor = self.cls if validate_db_data else self.cls.construct

        return [
            constructor(
                from_db=True, _doc=doc, **doc.to_dict(), _key=doc.reference.path
            )
            for doc in docs
        ]

    """START OF EDITED PART"""

    def model_from_dict(
        self,
        *,
        d: T.Optional[dict],
        key: str = None,
        doc: magicdb.DocumentSnapshot = None,
        create: bool = False,
        validate_db_data: bool = True,
    ):

        _key = self.key if not key else key

        if not d:
            return None if not create else self.cls(_key.split("/")[-1])

        constructor = self.cls if validate_db_data else self.cls.construct

        return constructor(
            from_db=True,
            _doc=doc,
            **d,
            _key=_key,
        )

    def d_from_firestore(self, **kwargs) -> T.Optional[dict]:
        with safe_span(self.make_safe_span_label("get")):
            doc = self.firebase_query.get(**kwargs)

        if type(doc) != DocumentSnapshot:
            return None

        with safe_span("to_dict"):
            return doc.to_dict()

    def model_mapping_from_firestore(
        self, refs: List[magicdb.DocumentReference], **kwargs
    ) -> T.Dict[str, dict]:
        with safe_span(self.make_safe_span_label("get_all")):
            docs = list(db.conn.get_all(refs, **kwargs))
        with safe_span(self.make_safe_span_label(f"to_dict {len(docs)=}")):
            return {doc.reference.path: doc.to_dict() for doc in docs}

    def model_mapping_from_redis(
        self, keys: List[str]
    ) -> T.Optional[T.Dict[str, dict]]:
        if not self.r:
            return None
        with safe_span(self.make_safe_span_label(f"*mget-redis*{len(keys)=}")):
            raw_strs = self.r.mget(keys)
        mapping = {}
        with safe_span(self.make_safe_span_label("*ujson.loads-from mget*")):
            for key, raw_str in zip(keys, raw_strs):
                value = None if raw_str is None else ujson.loads(raw_str)
                mapping[key] = value
        return mapping

    def d_from_redis(self) -> T.Optional[dict]:
        if not self.r:
            return None
        with safe_span(self.make_safe_span_label("*get-redis*")):
            raw_str = self.r.get(self.key)
        if not raw_str:
            return None

        with safe_span(self.make_safe_span_label("*ujson.loads*")):
            return ujson.loads(raw_str)

    def model_mapping_to_redis(self, mapping: T.Dict[str, dict]) -> None:
        if not self.r:
            return None

        json_mapping = {}
        to_delete = []
        with safe_span(
            self.make_safe_span_label(f"construct-from-getall-miss:{self.cls.__name__}")
        ):
            for key, d in mapping.items():
                if not d:
                    to_delete.append(key)
                    continue
                model = self.cls.construct(**d)
                value = model.json(exclude_unset=True, exclude={"id"})
                json_mapping[key] = value
        with safe_span(self.make_safe_span_label("mset-redis-from-miss")):
            self.r.mset(json_mapping)
            if self.r_ttl_secs:
                with self.r.pipeline() as pipe:
                    for key in json_mapping.keys():
                        pipe.expire(key, self.r_ttl_secs)
                    pipe.execute()
        if to_delete:
            with safe_span(self.make_safe_span_label("delete-redis-from-miss")):
                self.r.delete(*to_delete)

    def d_to_redis(self, d: dict) -> None:
        if not self.r or not d:
            return
        with safe_span(
            self.make_safe_span_label(f"construct-from-miss:{self.cls.__name__}")
        ):
            j = self.cls.construct(**d, _key=self.key).json(
                exclude_unset=True, exclude={"id", "_key"}
            )
        with safe_span(self.make_safe_span_label("set-redis-from-miss")):
            if self.r_ttl_secs:
                self.r.setex(self.key, self.r_ttl_secs, j)
            else:
                self.r.set(self.key, j)

    def get_mapping_from_db(self, ids: List[str], **kwargs) -> T.Dict[str, dict]:
        refs = [db.conn.collection(self.collection_path).document(id) for id in ids]
        keys = [ref.path for ref in refs]

        model_mapping: T.Dict[str, dict] = self.model_mapping_from_redis(keys=keys)

        # d_arr = self.d_arr_from_redis(keys=keys)
        if None in model_mapping.values():
            print("cache miss for get all")
            model_mapping = self.model_mapping_from_firestore(refs=refs, **kwargs)
            self.model_mapping_to_redis(mapping=model_mapping)
        return model_mapping

    def get_d_from_db(self, **kwargs) -> T.Optional[dict]:
        d = self.d_from_redis()
        if not d:
            # cache miss so get from firestore
            print("cache miss")
            d = self.d_from_firestore(**kwargs)
            self.d_to_redis(d=d)
        return d

    def get_internal(self, create=False, validate_db_data=True, **kwargs):
        d = self.get_d_from_db(**kwargs)

        return self.model_from_dict(
            d=d, create=create, validate_db_data=validate_db_data
        )

    @magic_async
    def get(self, id=None, create=False, validate_db_data=True, **kwargs):
        if id:
            self.firebase_query = self.firebase_query.document(id)

        return self.get_internal(
            create=create, validate_db_data=validate_db_data, **kwargs
        )

    @magic_async
    def get_all(self, ids: List[str], validate_db_data=True, **kwargs) -> list:
        mapping = self.get_mapping_from_db(ids=ids, **kwargs)

        with safe_span(self.make_safe_span_label("get_all-creating models")):
            models = [
                self.model_from_dict(
                    d=d, key=key, doc=None, validate_db_data=validate_db_data
                )
                for key, d in mapping.items()
            ]
        return self.sort_models_by_id_order(models, ids)

    @staticmethod
    def sort_models_by_id_order(models, ids):
        """Maintain order of return array"""
        sorted_models = [None] * len(ids)
        sort_key = {id: index for index, id in enumerate(ids)}
        for model in models:
            sorted_models[sort_key[model.id]] = model
        return sorted_models

    """REDIS"""

    @property
    def r(self) -> T.Optional[redis.Redis]:
        return self.cls.get_redis()

    @property
    def r_ttl_secs(self) -> T.Optional[int]:
        return self.cls.get_redis_ttl_secs()

    """END OF EDITED PART"""

    """Subcollections"""

    def collections(self):
        return self.firebase_query.collections()

    def collections_names(self):
        return [coll.id for coll in self.firebase_query.collections()]



from pydantic import BaseModel, Field, PrivateAttr
import magicdb
from magicdb import db
from magicdb.Models import MagicModelSpeed


class MagicFieldsCreator(BaseModel):
    id: str = None
    _key: str = PrivateAttr(None)
    _ref: magicdb.DocumentReference = PrivateAttr(None)
    _doc: magicdb.DocumentSnapshot = PrivateAttr(None)
    _parent: MagicModelSpeed = PrivateAttr(None)

    @staticmethod
    def make_magic_fields(magic_model: MagicModelSpeed, data: dict):
        MagicFieldsCreator()._make_magic_fields(magic_model, data)

    def _make_magic_fields(self, magic_model: MagicModelSpeed, data: dict):
        for field in self.__annotations__.keys():
            setattr(self, field, data.get(field, None))
        if self._key:
            self.fields_from_key()
        if self._ref:
            self.fields_from_ref()
        else:
            self.fields_from_id_parent_nothing(magic_model.collection_name)

        magic_model.id = self.id
        magic_model._key = self._key
        magic_model._ref = self._ref
        magic_model._doc = self._doc
        magic_model._parent = self._parent

    def fields_from_key(self):
        self._ref = db.conn.document(self._key)
        self.id = self.id_from_ref(self._ref)

    def fields_from_ref(self):
        self._key = self.key_from_ref(self._ref)
        self.id = self.id_from_ref(self._ref)

    def fields_from_id_parent_nothing(self, collection_name: str):
        base_ref = db.conn if not self._parent else self._parent._ref
        coll_ref = base_ref.collection(collection_name)
        self._ref = coll_ref.document() if not self.id else coll_ref.document(self.id)
        self.fields_from_ref()

    @staticmethod
    def id_from_ref(ref):
        return ref.path.split("/")[-1]

    @staticmethod
    def key_from_ref(ref):
        return ref.path

    """SETTING AND GETTING"""

    @staticmethod
    def set_id(magic_model: MagicModelSpeed, id: str):
        MagicFieldsCreator()._make_magic_fields(magic_model, {"id": id})

    @staticmethod
    def set_key(magic_model: MagicModelSpeed, key: str):
        MagicFieldsCreator()._make_magic_fields(magic_model, {"_key": key})

    @staticmethod
    def set_ref(magic_model: MagicModelSpeed, ref: magicdb.DocumentReference):
        MagicFieldsCreator()._make_magic_fields(magic_model, {"_ref": ref})

    @staticmethod
    def set_parent(magic_model: MagicModelSpeed, parent: MagicModelSpeed):
        MagicFieldsCreator()._make_magic_fields(magic_model, {"_parent": parent})



from datetime import datetime

from pydantic import Field

from magicdb.Models import MagicModel


class DateModel(MagicModel):
    created_at: datetime = Field(default_factory=datetime.utcnow)
    last_updated: datetime = Field(default_factory=datetime.utcnow)

    def update(self, last_updated=None, *args, **kwargs):
        self.last_updated = last_updated or datetime.utcnow()
        return super().update(*args, **kwargs)



import typing as T
import magicdb
from magicdb import db
from magicdb.Models import MagicModelSpeed


class FieldMaker:
    def __init__(
        self,
        *,
        _key: str = None,
        _ref: magicdb.DocumentReference = None,
        _collection_name: str,
        id: str = None,
        _parent: MagicModelSpeed = None,
        **kwargs,  # this is so that all fields can be passed and I only keep ones that matter
    ):
        self._key = _key
        self._ref = _ref
        self.collection_name = _collection_name
        self.id = id
        self._parent = _parent

        self.make_magic_fields_from_scratch()

    @classmethod
    def from_model(cls, magic_model: MagicModelSpeed):
        return cls(
            _key=magic_model._key,
            _ref=magic_model._ref,
            _collection_name=magic_model.get_collection_name(),
            id=magic_model.id,
            _parent=magic_model._parent,
        )

    def make_magic_fields_from_scratch(self):
        if self._key:
            self.from_key()
        elif self._ref:
            self.from_ref()
        else:
            self.from_id_parent_collection_name()

    def set_model_fields(
        self, magic_model: MagicModelSpeed, fields_to_exclude: T.Set[str] = None
    ):
        if not fields_to_exclude:
            fields_to_exclude = set()

        if "id" not in fields_to_exclude:
            magic_model._id = self.id
        if "key" not in fields_to_exclude:
            magic_model._key = self._key
        if "ref" not in fields_to_exclude:
            magic_model._ref = self._ref
        if "parent" not in fields_to_exclude:
            magic_model._parent = self._parent

    def from_ref(self):
        self.key_from_ref()
        self.id_from_ref()

    def from_key(self):
        self.ref_from_key()
        self.from_ref()

    def from_id_parent_collection_name(self):
        base_ref = db.conn if not self._parent else self._parent._ref
        coll_ref = base_ref.collection(self.collection_name)
        self._ref = coll_ref.document() if not self.id else coll_ref.document(self.id)
        self.from_ref()

    def id_from_key(self):
        self.id = self._key.split("/")[-1]

    def id_from_ref(self):
        self.id = self._ref.path.split("/")[-1]

    def ref_from_key(self):
        self._ref = db.conn.document(self._key)

    def key_from_ref(self):
        self._key = self._ref.path



from typing import Union, List
from magicdb import db, ASCENDING, DESCENDING, DocumentSnapshot
from magicdb.utils.Serverless.span import safe_span
from magicdb.utils.async_helpers import threadpool_asyncio, threadpool, magic_async

import time


class Query:
    def __init__(self, cls):
        from magicdb.Models.MagicModel import MagicModel

        self.cls: MagicModel = cls
        self.collection_path = self.cls.get_collection_name()
        self.firebase_query = self.create_query()
        self.query_inputs: List[str] = []

    def create_query(self):
        self.firebase_query = db.conn.collection(self.collection_path)
        return self.firebase_query

    def parent(self, parent_or_path):
        from magicdb.Models.MagicModel import MagicModel

        parent_path = (
            parent_or_path
            if not issubclass(type(parent_or_path), MagicModel)
            else parent_or_path.key
        )

        self.collection_path = parent_path + "/" + self.collection_path
        self.create_query()
        return self

    def document(self, document_path):
        self.firebase_query = self.firebase_query.document(document_path)
        return self

    def collection_group(self):
        self.firebase_query = db.conn.collection_group(self.cls.get_collection_name())
        return self

    def collection(self, collection_path):
        self.firebase_query = self.firebase_query.collection(collection_path)
        return self

    """Wrapping Firestore abilities."""

    def where(self, field, action, value):
        self.firebase_query = self.firebase_query.where(field, action, value)
        self.query_inputs.append(f"{field}, {action}, {value}")
        return self

    def order_by(self, field, direction=None, **kwargs):
        if direction:
            kwargs["direction"] = direction
        if kwargs.get("direction", None) == "asc":
            kwargs["direction"] = ASCENDING
        if kwargs.get("direction", None) == "desc":
            kwargs["direction"] = DESCENDING
        self.firebase_query = self.firebase_query.order_by(field, **kwargs)
        return self

    def start_at(self, fields):
        self.firebase_query = self.firebase_query.start_at(fields)
        return self

    def start_after_id(self, last_doc_id: str = None):
        if not last_doc_id:
            return self
        key = self.cls.construct(id=last_doc_id).key
        snapshot = db.conn.document(key).get()
        self.firebase_query = self.firebase_query.start_after(snapshot)
        return self

    def start_after(self, fields):
        self.firebase_query = self.firebase_query.start_after(fields)
        return self

    def end_at(self, fields):
        self.firebase_query = self.firebase_query.end_at(fields)
        return self

    def end_before(self, fields):
        self.firebase_query = self.firebase_query.end_before(fields)
        return self

    def limit(self, limit):
        if not limit or limit == -1:
            return self
        self.firebase_query = self.firebase_query.limit(limit)
        return self

    def get_path_of_query(self):
        if getattr(self.firebase_query, "_path", None):
            return "/".join(self.firebase_query._path)
        if getattr(self.firebase_query, "_parent", None):
            return "/".join(self.firebase_query._parent._path)
        else:
            return self.collection_path

    """Querying and creating of the models."""

    def make_safe_span_label(self, query_type):
        return f"{query_type}-{self.get_path_of_query()}-{str(self.query_inputs)}"

    @magic_async
    def stream(self, validate_db_data=True, **kwargs):
        """Currently defaults to validating the db data, but you can turn this off with the validate_db_data param"""
        with safe_span(self.make_safe_span_label("stream")):
            docs = list(self.firebase_query.stream(**kwargs))

        constructor = self.cls if validate_db_data else self.cls.construct

        return [
            constructor(from_db=True, doc=doc, **doc.to_dict(), key=doc.reference.path)
            for doc in docs
        ]

    @threadpool_asyncio
    def stream_async(self, validate_db_data=True, **kwargs):
        with safe_span(self.make_safe_span_label("stream")):
            docs = list(self.firebase_query.stream(**kwargs))

        constructor = self.cls if validate_db_data else self.cls.construct

        return [
            constructor(from_db=True, doc=doc, **doc.to_dict(), key=doc.reference.path)
            for doc in docs
        ]

    @magic_async
    def get(self, id=None, create=False, validate_db_data=True, **kwargs):
        if id:
            self.firebase_query = self.firebase_query.document(id)

        with safe_span(self.make_safe_span_label("get")):
            doc = self.firebase_query.get(**kwargs)

        if type(doc) != DocumentSnapshot:
            return None

        d = doc.to_dict()
        if not d:
            return None if not create else self.cls(id=id)

        constructor = self.cls if validate_db_data else self.cls.construct

        return constructor(
            from_db=True,
            doc=doc,
            **doc.to_dict(),
            key=doc.reference.path,
        )

    @staticmethod
    def sort_models_by_id_order(models, ids):
        """Maintain order of return array"""
        sorted_models = [None] * len(ids)
        sort_key = {id: index for index, id in enumerate(ids)}
        for model in models:
            sorted_models[sort_key[model.id]] = model
        return sorted_models

    @magic_async
    def get_all(self, ids, validate_db_data=True, **kwargs):
        doc_refs = [db.conn.collection(self.collection_path).document(id) for id in ids]

        with safe_span(self.make_safe_span_label("get_all")):
            docs = list(db.conn.get_all(doc_refs, **kwargs))

        constructor = self.cls if validate_db_data else self.cls.construct

        models = [
            constructor(
                from_db=True,
                doc=doc,
                **doc.to_dict(),
                key=doc.reference.path,
            )
            for doc in docs
        ]
        return self.sort_models_by_id_order(models, ids)

    @threadpool_asyncio
    def get_all_async(self, ids, validate_db_data=True, **kwargs):
        return self.get_all(ids, validate_db_data, **kwargs)

    """Subcollections"""

    def collections(self):
        return self.firebase_query.collections()

    def collections_names(self):
        return [coll.id for coll in self.firebase_query.collections()]



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['magicdb',
 'magicdb.FirestoreWrappers',
 'magicdb.Models',
 'magicdb.Queries',
 'magicdb.Timing',
 'magicdb.database',
 'magicdb.utils',
 'magicdb.utils.Serverless']

package_data = \
{'': ['*']}

install_requires = \
['firebase-admin>=4.4.0,<5.0.0',
 'glom>=20.8.0,<21.0.0',
 'grpcio>=1.34.0,<2.0.0',
 'pydantic>=1.7.0,<2.0.0',
 'python-slugify>=4.0.1,<5.0.0',
 'redis>=3.5.3,<4.0.0',
 'ujson>=4.0.2,<5.0.0']

setup_kwargs = {
    'name': 'magicdb',
    'version': '0.2.122',
    'description': '',
    'long_description': '# MagicDB\nA fully typed Firestore ORM for python -- the easiest way to store data.\n\nMagicDB inherets from Pydantic, so you get all the power of Pydantic models with the functionality of Firestore: https://pydantic-docs.helpmanual.io/.\n\n## Instalation\n```\npip install magicdb\n```\n\n## Initialize the DB\nMagicDB is initialized via a Firestore service account json which you download from your Firebase console.\nOnce you have the json, you must tell MagicDB where it is, either by 1) setting the `GOOGLE_APPLICATION_CREDENTIALS` environment variable to the json path, or by 2) calling magicdb.connect with the path:\n\n```python\n# 1)\n# You can set the env variable from the terminal too: export GOOGLE_APPLICATION_CREDENTIALS="path/to/my-service-account.json"\nimport os\nos.environ["GOOGLE_APPLICATION_CREDENTIALS"] = "path/to/my-service-account.json"\n\n# OR\n\n# 2)\nimport magicdb\nmagicdb.connect(from_file="path/to/my-service-account.json")\n```\n\n## Example\n```python\nfrom magicdb.Models import MagicModel\n\nclass Salesman(MagicModel):\n    name: str = None\n    company: str = None\n\ns = Salesman()\ns.name = \'Jim\'\ns.save()\n\n# Get Salesman\ns = Salesman.collection.get(s.id)\nprint(s.name) # Jim\n```\n\n## Fields\nUse any type [mypy](http://mypy-lang.org/) will accept!\n\n#### Fields Example\n```python\nfrom datetime import datetime\n\nclass Manager(MagicModel):\n\tname: str\n\tage: int\n\tcompany: str = \'Dunder Mifflin\'\n\tstartedWorkingAt: datetime = None\n\n# m = Manager(name=\'Michael Scott\', age=44)  # you must pass in the required fields on initializing the object.\nm.age = 45\nm.save()  # Success! New doc in collection "manager" as: { name: Michael Scott, age: 45, company: Dunder Mifflin }\n\nm = Manager(name=\'Dwight Schrute\') # Exception since age is required but not given\n```\n\nYou can also add other Objects as a field.\n\n### NestedModel Example\n```python\nclass Dog(MagicModel):\n\tage: int\n\towner: Manager\n\ndog = Dog()\ndog.age = 3\ndog.owner = Manager(name=\'Robert California\', age=59)\ndog.save()\nprint(dog)\n\n```\n\n\n## Collections\nThe collection name for a class defaults to the class\' name in lowercase. To set the collection name, use the `Meta` class.\n\n### Meta Example\n\n```python\nclass Student(MagicModel):\n\tname: str = None\n\tschool: str = \'UPenn\'\n\n\tclass Meta:\n\t\tcollection_name = \'students\'\n\n\ns = Student(name=\'Amy Gutman\')\ns.save()  # creates a new document in the "students" collection\nprint(s)  # name=\'Amy Gutman\' school=\'UPenn\'\n```\n\nYou can also inheret classes.\n\n### Inheritance Example\n```python\nclass ExchangeStudent(Student):\n\toriginalCountry: str\n\n\tclass Meta:\n\t\tcollection_name = \'exchangeStudents\'\n\ne = ExchangeStudent(originalCountry=\'UK\')\nprint(e.school)  # UPenn\ne.save()\nprint(e)  # name=None school=\'UPenn\' originalCountry=\'UK\'\n```\n\n## Queries\nYou can make queries with the same syntax you would using the Python firebase-admin SDK. But FireORM returns the objects.\n\n### Queries Example\n```python\n\ne = ExchangeStudent(originalCountry=\'UK\')\nprint(e.school)  # UPenn\ne.save()\nprint(e)  # name=None school=\'UPenn\' originalCountry=\'UK\'\n\nmanagers = Manager.collection.where(\'name\', \'==\', \'Michael Scott\').limit(1).stream()\nprint(managers) # [Manager(name=\'Michael Scott\', age=45, company=\'Dunder Mifflin\', startedWorkingAt=None)]\nprint(managers[0].id)\nmanager = Manager.collection.get(\'0mIWZ8FfgQzBanCllqsV\')\nprint(manager) # name=\'Michael Scott\' age=45 company=\'Dunder Mifflin\' startedWorkingAt=None\n```\n',
    'author': 'Jeremy Berman',
    'author_email': 'jerber@sas.upenn.edu',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/jerber/magicdb_new',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



from magicdb.Models.MagicCache import MagicCache
from magicdb.Models.MagicModel import MagicModel
from magicdb.Models.MagicModelSpeed import MagicModelSpeed
from magicdb.Models.DateModel import DateModel
from magicdb.Models.DateModelSpeed import DateModelSpeed
from magicdb.Models.FrontendParserModel import FrontendParserModel
from magicdb.Models.RemoveFields import RemoveFieldsMeta
from magicdb.Models.RemoveFieldsSpeed import RemoveFieldsMetaSpeed



import os
from contextlib import contextmanager

SHOULD_USE_SPAN = False
use_span_env = os.getenv("USE_SPAN", "true").lower()
use_span = "true" in use_span_env or "1" in use_span_env

try:
    if use_span:
        from serverless_sdk import span

        SHOULD_USE_SPAN = True
except ModuleNotFoundError as e:
    SHOULD_USE_SPAN = False

print("SHOULD_USE_SPAN", SHOULD_USE_SPAN)

import time


@contextmanager
def print_span(label: str):
    start = time.time()
    yield
    time_took = time.time() - start
    print(f"{label} took {time_took*1_000:.5f} ms.")


@contextmanager
def safe_span(label, use=True):
    if not SHOULD_USE_SPAN or not use:
        with print_span(label):
            yield
    else:
        with span(label):
            yield
